/*=========================================*\

  Jeudi 7 Janvier 1999
  
  Arash Habibi
  
  Timer.c
  
  Les structures de donnees pour les 
  timers. 
  
  On distingue le timer systeme qui 
  est regle sur l'horloge du systeme et
  tous les timers de notre application 
  qui s'en distinguent par des offsets
  et qui sont remis a jour a chaque fois
  qu'un timeout est expire ou qu'un 
  nouveau timeout est introduit. 
  
\*=========================================*/


#include "Timer.h"

AppliTimer_t *FirstTimer;    /* Le timer le plus urgent   */
int TimerPipe;               /* Le tube ou on envoie le   */
			     /* le numero du timer expire */


void InitSysTimer(struct itimerval *SysTimer)
/*=========================================*\
  Initialise la variable SysTimer a une 
  valeur nulle (qui desarmerait le timer
  s'il etait effectivement applique au 
  timer du systeme.)
\*=========================================*/
{ 
  SysTimer->it_interval.tv_sec  = 0; 
  SysTimer->it_interval.tv_usec = 0;
  
  SysTimer->it_value.tv_sec  = 0; 
  SysTimer->it_value.tv_usec = 0; 
}


void StopSysTimer()
/*=========================================*\
  Desarme le timer systeme de facon a ce que
  plus aucun signal SIGALRM ne sera delivre. 
\*=========================================*/
{
  struct itimerval SysTimer;
  
  SysTimer.it_interval.tv_sec  = 0; 
  SysTimer.it_interval.tv_usec = 0;
  
  SysTimer.it_value.tv_sec  = 0; 
  SysTimer.it_value.tv_usec = 0; 

  setitimer(ITIMER_REAL, &SysTimer, NULL); 
}


void SetSysTimerValue(float duree)
/*=========================================*\
  Affecte duree a la valeur du timer du 
  systeme. 
\*=========================================*/
{
  float value; 
  struct itimerval SysTimer; 
  long sec, microsec; 

  sec = (long)duree; 
  microsec = (long)(1000000*(duree - (long)duree)); 

  InitSysTimer(&SysTimer); 
  
  SysTimer.it_value.tv_sec  = (long)duree; 
  SysTimer.it_value.tv_usec = 
    (long)(1000000*(duree - (long)duree)); 
  
  setitimer(ITIMER_REAL, &SysTimer, NULL); 
}


float GetSysTimerValue()
/*=========================================*\
   Donne (en secondes) la duree jusqu'a 
   expiration du timeout et de l'emission
   d'un signal SIGALRM.
\*=========================================*/
{
  float value;
  struct itimerval SysTimer; 
  
  getitimer(ITIMER_REAL, &SysTimer);
  
  value = (float)(SysTimer.it_value.tv_sec
    + SysTimer.it_value.tv_usec/1000000.);

  return(value);
}


AppliTimer_t* NewAppliTimer(int identite, float duree)
/*=========================================*\
  Alloue un nouveau timer appli et initialise
  sa duree a "duree". 
\*=========================================*/
{ 
  AppliTimer_t *timer; 

  timer = (AppliTimer_t*)malloc(sizeof(AppliTimer_t));

  if(timer==NULL)
    {fprintf(stderr,"NewTimer : malloc a echoue.\n");
    exit(1);}

  else
    {
      timer->identite = identite; 
      timer->duree    = duree; 
      timer->next     = NULL; 
    }
  return(timer);
}


void InitTimers(int TPipe)
/*=========================================*\
  Initialise la liste des timers de notre
  application. Au depart cette liste est 
  vide. 
\*=========================================*/
{ 
  struct itimerval SysTimer; 
  struct sigaction p_action; 

  InitSysTimer(&SysTimer);

  p_action.sa_handler = &TimeoutHandler;
  sigaction(SIGALRM, &p_action, NULL); 
  
  FirstTimer=NULL; 
  
  TimerPipe = TPipe; 
}


void UpdateAllAppliTimers()
/*=========================================*\
  Lit la valeur de SysTimer. SysTimer a ete
  arme avec le premier timer de la liste.
  Faire la difference entre SysTimer et 
  ce temps et remplacer toutes les durees
  de tous les timers par la valeur prect
  - FirstTimer->duree + SysTimer. 
\*=========================================*/
{
  float offset; 
  AppliTimer_t *timer; 
  
  if(FirstTimer)
    {
      offset = FirstTimer->duree - GetSysTimerValue(); 
      
      FOR_ALL_TIMERS_IN_LIST(timer) 
	timer->duree -= offset; 
    }
}



void RemoveFirstAppliTimer()
/*=========================================*\
  Remplace le premier timer par le deuxieme
  et elimine le premier timer. 
\*=========================================*/
{
  AppliTimer_t *Morituri;

  Morituri = FirstTimer; 
  FirstTimer = FirstTimer->next; 
  free(Morituri);
}


void InsertFirstAppliTimer(AppliTimer_t *timer)
/*=========================================*\
  timer sera place au premier rang.
\*=========================================*/
{
  struct itimerval *Morituri;

  timer->next = FirstTimer; 
  FirstTimer  = timer; 
}


void InsertAppliTimer(AppliTimer_t *timer, AppliTimer_t *prec)
/*=========================================*\
  timer sera insere juste devant prec. 
  En d'autre termes, timer doit etre un peu
  plus urgent que prec. 
\*=========================================*/
{
  timer->next = prec->next; 
  prec->next  = timer; 
}


void RemoveAppliTimer(AppliTimer_t *RemovedTimer)
/*=========================================*\
  RemovedTimer sera elimine de la liste des 
  timers. 
\*=========================================*/
{
  AppliTimer_t *timer, *prev_timer; 

  if(RemovedTimer==FirstTimer)
    RemoveFirstAppliTimer();

  else
    {
      FOR_ALL_TIMERS_IN_LIST(timer) 
	if(timer->next == RemovedTimer)
	  prev_timer = timer; 
      
      /* prev_timer est le timer juste avant RemovedTimer */
      
      prev_timer->next = RemovedTimer->next; 
    }

  free(RemovedTimer); 
}


void ClassifyAppliTimerInList(AppliTimer_t *NewTimer)
/*=========================================*\
  Place le nouveau timer derriere ceux qui 
  sont plus urgents et devant ceux qui le 
  sont moins. 
\*=========================================*/
{
  AppliTimer_t *timer; 

/* ----------Insertion au premier rang ----------- */

  if((FirstTimer==NULL)||
     (NewTimer->duree < FirstTimer->duree))
    InsertFirstAppliTimer(NewTimer);

  else
    {
      FOR_ALL_TIMERS_IN_LIST(timer) 
	{
/* --------- Insertion au dernier rang ------------*/       

	if(((timer->next==NULL)&&    
	    (NewTimer->duree > timer->duree)) ||

/* -------------Les autres cas --------------------*/

	   ((timer->next!=NULL)&&
	    (NewTimer->duree >= timer->duree)&&
	    (NewTimer->duree <  timer->next->duree)))

	  InsertAppliTimer(NewTimer, timer); 
	}
    }
}


void CheckTimers(char *comment)
/*=========================================*\
  Affiche a l'ecran tous les timers armes 
  avec leurs caracteristiques. 
\*=========================================*/
{
  AppliTimer_t *timer; 

  fprintf(stderr,"\nCHECK-TIMERS : %s\n", comment);

  FOR_ALL_TIMERS_IN_LIST(timer) 
    fprintf(stderr,"Timer %d \t a %f secondes de l'expiration.\n",
	    timer->identite, timer->duree);

  fprintf(stderr,"\n");
}


void StartATimer(int identite, float duree)
/*=========================================*\
  Remettre a jour toutes les durees de tous 
  les temporisateurs. 
  Si duree est plus petite que toutes les 
  autres, la mettre en tete et changer la 
  valeur de SysTimer. 
  Sinon classer le nouveau timer parmi les 
  autres dans l'ordre des durees. 
\*=========================================*/
{
  AppliTimer_t *timer, *NewTimer; 
  boolean impossible=FAUX; 

/*--------- Voir si le timer n'existe pas deja ----------*/

  FOR_ALL_TIMERS_IN_LIST(timer)
    if(timer->identite == identite)
      {fprintf(stderr,"Le timer %d existe deja.\n",identite);
      impossible=VRAI;}

/*--------- Valeurs de parametre incorrectes ------------*/

  if((!impossible)&&(identite<=0))
      {fprintf(stderr,"StartATimer : L'identite doit etre strictement positive.\n");
      fprintf(stderr,"La valeur de %d n'est pas autorisee.\n", identite);
      impossible=VRAI;}

  if((!impossible)&&(duree<=0))
      {fprintf(stderr,"StartATimer : La duree doit etre strictement positive.\n");
      fprintf(stderr,"La valeur %f donnee pour le timer %d n'est pas autorisee.\n",
	      duree, identite);
      impossible=VRAI;}

/*--------- Allocation et insertion dans la liste---------*/

  if(!impossible)
    {
      NewTimer = NewAppliTimer(identite, duree); 
     
      UpdateAllAppliTimers();
      
      if((FirstTimer==NULL)||
	 ((FirstTimer!=NULL)&&(duree < FirstTimer->duree)))	
	SetSysTimerValue(duree);
      
      ClassifyAppliTimerInList(NewTimer);
    }
}



void StopTimer(int identite)
/*=========================================*\
  Enleve le timer identite de la liste. 
\*=========================================*/
{
  AppliTimer_t *timer; 
  boolean trouve=FAUX; 

  if(FirstTimer->identite == identite)
    {
      if(FirstTimer->next!=NULL)
	SetSysTimerValue(FirstTimer->next->duree);
      else
	StopSysTimer();

      RemoveFirstAppliTimer();
      trouve=VRAI; 
    }

  else
    FOR_ALL_TIMERS_IN_LIST(timer) 
      if(timer->identite == identite)
	{RemoveAppliTimer(timer);  trouve=VRAI;}

  if(trouve==FAUX)
    fprintf(stderr,"StopTimer : le timer %d n'existe pas.\n",
	    identite);
}



int GetNextExpiredTimerID(int PipeDesc)
/*=========================================*\
  Tant qu'il y a un nombre dans le tube 
  decrit par PipeDesc, cette fonction rend
  ce nombre, sinon, il rend PLUS_RIEN. 
\*=========================================*/
{
  int res, TimerID=0; 

  res = read(PipeDesc, &TimerID, sizeof(int)); 

  if(res==CRASH) 
    {perror("GetNextExpiredTimerID : read");
    return(CRASH);}

  else if(res==PLUS_RIEN) return(PLUS_RIEN); 
  
  else return(TimerID);
}


void TimeoutHandler()
/*=========================================*\
  Il faut enlever le permier timer
  de la liste. Mettre son id dans TimerPipe
  et diminuer tous les temporisations des 
  autres Timer_t de la duree initiale du 
  timer qu'on vient d'ecarter. 
  Rearmer le temporisateur avec la duree 
  du nouveau premier timer.  
\*=========================================*/
{
  int res; 
  float Duree; 

  fprintf(stderr,"Expiration du timeout numero %d\n",
	  FirstTimer->identite);

  res = write(TimerPipe, &(FirstTimer->identite), sizeof(int)); 
  if(res==-1) perror("ReactionAExpiration write");

  UpdateAllAppliTimers();

  if(FirstTimer->next!=NULL)
    SetSysTimerValue(FirstTimer->next->duree);
  else
    StopSysTimer();

  RemoveFirstAppliTimer();
}



void main()
/*=========================================*\
\*=========================================*/
{
  int *Tubes, tube_lect, tube_ecri, res, i; 

  res = pipe(Tubes);
  if(res==-1) {perror("tube"); exit(1);}
  else{tube_lect = Tubes[0]; 
       tube_ecri = Tubes[1];}

  InitTimers(tube_ecri);

  StartATimer(5, 2.51);
  CheckTimers("Apres la premiere allocation");

  StartATimer(3, 5);
  CheckTimers("Apres la deuxieme allocation");

  /*sleep(1);
  UpdateAllAppliTimers();
  CheckTimers("Apres le sleep");*/

  for(i=0; i<2; i++)
    {
      pause();
      CheckTimers("Apres expiration");
    }

  res=TOUT_BAIGNE;
  while((res!=CRASH)&&(res!=PLUS_RIEN))
    {
      res=GetNextExpiredTimerID(tube_lect);

      if(res==CRASH)
	fprintf(stderr,"Probleme pour lire le tube.\n");
      else if(res==PLUS_RIEN)
	fprintf(stderr, "C'est tout comme timeout.\n");
      else
	fprintf(stderr, "Timeout numero %d\n");
    }
}
