
/*================================================*\

	Lundi 23 novembre 1998

	Arash HABIBI

	Liaison.c

        Le module principal de la couche liaison.

\*================================================*/


#include "Liaison.h"


int ReactionArriveeCommande(int *DataAppli)
/*================================================*\
  Va chercher la commande de la couche application 
  et rend le diagnostique. 
\*================================================*/
{
  int commande =1;		/* TOUT_BAIGNE */
  int descripteur=-1;		/* a priori rien */

  if(OrigineCmdAppli(&commande, &descripteur)==-1)
    return(CRASH);
  
  if(commande==QUIT) return(TERMINE);
  
  else if(commande==ENVOI) 
    *DataAppli = descripteur; 
  
  else /* rien a faire */ ;
  
  return(TOUT_BAIGNE);
}


int ReactionArriveePaquet()
/*================================================*\
   Lit ce qu'il y a a lire venant de l'appli
   le met dans une trame et l'envoie a la couche
   physique. Si tout s'est bien passe, on rend 
   TOUT_BAIGNE, si l'emission est terminee, alors
   on rend EMISSION_FINIE et sinon on rend CRASH.
\*================================================*/
{
    int diagnostic=TOUT_BAIGNE;
    int taille_paquet=0, taille_trame=0; 
    Paquet_t paquet; 
    Trame_t trame; 
    
    diagnostic = OrigineApplication(&paquet, 
				    &taille_paquet);
	  
    if((diagnostic!=CRASH)&&(diagnostic!=EMISSION_FINIE))
       {
	 EncapsulerPaquet(paquet,  taille_paquet, 
			  &trame , &taille_trame);
	 
	 diagnostic = VersCouchePhysique(&trame, 
					 taille_trame);
       }
    return(diagnostic);		  
}


int ReactionArriveeTrame()
/*================================================*\
   Lit ce qu'il y a a lire venant de la couche phys
   extrait le paquet d'information et le delivre 
   a l'application. Si tout s'est bien passe, on rend 
   TOUT_BAIGNE, sinon on rend CRASH.
\*================================================*/
{
   int diagnostic = TOUT_BAIGNE; 
   int taille_paquet, taille_trame; 
   Paquet_t paquet; 
   Trame_t trame; 
     
   diagnostic = OrigineCouchePhysique (&trame, 
				       &taille_trame);
   if(diagnostic!=CRASH)
      {
	DesEncapsulerPaquet (trame, taille_trame, 
			     &paquet, &taille_paquet);
	
	diagnostic = VersApplication (&paquet, 
				      taille_paquet);
      }
   return(diagnostic);
}


/**/int ReactionTimeout(int *tube)
/**//*================================================*/
/**//*   La reaction au fait qu'un des timeout lances */
/**//*   est expire.                                  */
/**//*================================================*/
/**/{
/**/   int res; 
/**/   
/**/   res=GetNextExpiredTimerID(*tube);
/**/
/**/   if(res==CRASH)
/**/	{fprintf(stderr,"Probleme pour lire le tube des timeout.\n");
/**/	 return(CRASH);}
/**/   else
/**/	{fprintf(stderr, "Timeout numero %d expire\n", res);
/**/	 return(TOUT_BAIGNE);}
/**/}
/**/
/**/ void DemoTimeout()
/**/ /*================================================*/
/**/ /*   Une demo pour montrer comment on utilise les */
/**/ /*   timeout particuliers.                        */
/**/ /*================================================*/
/**/ {
/**/    CheckTimers("Avant d'avoir arme le premier timer");
/**/ /*---- Fonction qui permet de voir l'etat de tous les timeouts */
/**/
/**/    StartATimer(1, 2.0);
/**/ /*---- Amorcage du timeout d'identite 1 et de duree 2.0 sec ---*/
/**/    StartATimer(2, 2.0);
/**/ /*---- Amorcage du timeout d'identite 2 et de duree 2.0 sec ---*/
/**/    CheckTimers("Apres avoir arme le premier et le second");
/**/
/**/    Attendre(RIEN, RIEN, RIEN, RIEN, 1.0);
/**/ /*----- On attend une seconde avant de lancer le troisieme timeout */
/**/
/**/    StartATimer(5, 3.0);
/**/ /*---- Amorcage du timeout d'identite 5 et de duree 3.0 sec ---*/
/**/    CheckTimers("Apres avoir arme le troisieme timeout.");
/**/
/**/    StartATimer(1254, 10.26543); 
/**/    CheckTimers("Apres avoir arme le quatrieme timeout.");
/**/
/**/    StopTimer(1254);
/**/ /*---- Pour Desamorcer le timeout numero 1254-------------------*/
/**/ /*---- (Juste pour voir comment ca marche) ---------------------*/
/**/    CheckTimers("Apres avoir desarme le quatrieme timeout.");
/**/ }

	  		  
void ExecCoucheLiaison(int DataPhys, int CmdAppli, int TubeTimeout)
/*================================================*\

  Cette fonction est celle qui definit le protocole.
  Tel quel, le protocole est squelettique. 

  DataPhys est un descripteur par lequel viennet
      les donnees venant de la couche physique. 

  DataAppli est un descripteur par lequel viennent
      les donnees venant de l'application

  CmdAppli est un descripteur par lequel viennent
      les commandes venant de la couche application
      par exemple les demandes de connexion pour dire
      qu'il y a un nouveau fichier a envoyer. 
      
  TubeTimeout est un descripteur de tube par lequel
       viennent les numeros des timeouts expires. 

  Les evenements susceptibles de survenir sont
  le fait qu'il y ait quelque chose a lire sur 
  l'un des descripteurs, ... ou alors un timeout
  general. 

  Dans un premier temps on n'ecoute que CmdAppli, 
  DataPhys et TubeTimeout. Si l'application a demande 
  une emission alors seulement on commence a ecouter 
  DataAppli. 

  Evidemment ceci n'a rien a voir avec HDLC. Les paquets
  sont simplement mis dans les trames sans aucun calcul 
  et envoyes tel quel. Rien n'est garanti, surtout s'il 
  y a des erreurs. 
  			A vous maintenant ...

\*================================================*/
{
   int evenement=1, diagnostic=TOUT_BAIGNE; 
   int DataAppli=RIEN;		/* la source des paquets */

   DemoTimeout();

   while((diagnostic!=CRASH)&&
	 (diagnostic!=TERMINE))
     {
       evenement = Attendre(DataAppli, DataPhys, CmdAppli, TubeTimeout, 0.0);

       if(evenement==-1) diagnostic=CRASH; 
       else 
	 {
	   /*    Si evenement == (arrivee-paquet) ------------------*/
	   /* ou si evenement == (arrivee-paquet-ET-trame)--------- */
	   /* ou si evenement == (arrivee-paquet-ET-cmd) ---------- */
	   /* ou si evenement == (arrivee-paquet-ET-trame-ET-cmd)-- */
	   /* etc etc etc...                                        */
	   
/**/	   if((evenement==YaALireSur1___)||(evenement==YaALireSur12__)||
/**/	      (evenement==YaALireSur1_3_)||(evenement==YaALireSur123_)||
/**/	      (evenement==YaALireSur1__4)||(evenement==YaALireSur12_4)||
/**/	      (evenement==YaALireSur1_34)||(evenement==YaALireSur1234))
	     
	     {diagnostic = ReactionArriveePaquet();
	     if(diagnostic==EMISSION_FINIE) DataAppli=RIEN;}
	   
	   
	   /*   Si evenement == (arrivee-trame) ----------------*/
	   /*ou si evenement == (arrivee-paquet-ET-trame)-------*/
	   /*ou si evenement == (arrivee-paquet-ET-cmd)---------*/
	   /*ou si evenement == (arrivee-paquet-ET-cmd-ET-cmd)--*/
	   /* etc etc etc...                                    */
	   
/**/	   if((evenement==YaALireSur_2__)||(evenement==YaALireSur12__)||
/**/	      (evenement==YaALireSur_23_)||(evenement==YaALireSur123_)||
/**/	      (evenement==YaALireSur_2_4)||(evenement==YaALireSur12_4)||
/**/	      (evenement==YaALireSur_234)||(evenement==YaALireSur1234))

	     diagnostic = ReactionArriveeTrame();

	   
	   /*---Si evenement == (arrivee-d'une-commande) -----*/
	   /*--- etc etc etc ---------------------------------*/
	   
/**/	   if((evenement==YaALireSur__3_)||(evenement==YaALireSur1_3_)||
/**/	      (evenement==YaALireSur_23_)||(evenement==YaALireSur123_)||
/**/	      (evenement==YaALireSur__34)||(evenement==YaALireSur1_34)||
/**/	      (evenement==YaALireSur_234)||(evenement==YaALireSur1234))
	       
	       diagnostic = ReactionArriveeCommande(&DataAppli);
	       
	   
/**/	   /*---Si evenement == Un des timeout est expire ----*/
/**/	   /*--- etc etc etc ---------------------------------*/
/**/	   
/**/	   if((evenement==YaALireSur___4)||(evenement==YaALireSur1__4)||
/**/	      (evenement==YaALireSur_2_4)||(evenement==YaALireSur12_4)||
/**/	      (evenement==YaALireSur__34)||(evenement==YaALireSur1_34)||
/**/	      (evenement==YaALireSur_234)||(evenement==YaALireSur1234))

/**/	       diagnostic = ReactionTimeout(&TubeTimeout);
	 }
     }
   VersCmdAppli();		/* pour dire a l'appli que   */
				/* la reception est terminee */
   
   if(diagnostic==CRASH) fprintf(stderr,"Echange interrompu.\n");
   else /* TERMINE */    fprintf(stderr,"Echange termine.\n");
}
