/*=========================================*\

  Jeudi 7 Janvier 1999
  
  Arash Habibi
  
  Timer.c
  
  Les structures de donnees pour les 
  timers. 
  
  On distingue le timer systeme qui 
  est regle sur l'horloge du systeme et
  tous les timers de notre application 
  qui s'en distinguent par des offsets
  et qui sont remis a jour a chaque fois
  qu'un timeout est expire ou qu'un 
  nouveau timeout est introduit. 

  Les timers alloues par l'utilisateur sont
  classes dans une liste classee du timeout 
  le plus urgent au moins urgent. Le premier 
  de la liste est le timeout le plus urgent. 
  Le timer systeme correspond a ce premier 
  timeout. Donc quand arrive SIGALRM, on sait
  toujours que c'est le premier timeout qui 
  est expire. Dans ce cas, on retire le 
  premier element de la liste et on recommence. 
  
\*=========================================*/


#include "Timer.h"

AppliTimer_t *FirstTimer;    /* Le timer le plus urgent   */
struct sigaction p_action; 
int TimerPipe;               /* Le tube ou on envoie le   */

			     /* le numero du timer expire */
/*
#     #
##   ##    ##     ####   #    #     #    #    #   ####
# # # #   #  #   #       #   #      #    ##   #  #    #
#  #  #  #    #   ####   ####       #    # #  #  #
#     #  ######       #  #  #       #    #  # #  #  ###
#     #  #    #  #    #  #   #      #    #   ##  #    #
#     #  #    #   ####   #    #     #    #    #   ####
*/

void MaskSigAlarm()
/*=========================================*\
   Il est necessaire que les timeout ne 
   derangent pas le systeme quand il ne 
   peut pas les traiter. Donc il faut 
   pouvoir masquer les signaux SIGALRM.
\*=========================================*/
{ 
   sigset_t ens_sig; 
   
   sigemptyset(&ens_sig); 
   sigaddset(&ens_sig, SIGALRM); 
   sigprocmask(SIG_SETMASK, &ens_sig, NULL);
}


void UnmaskSigAlarm()
/*=========================================*\
   Pour enlever le masque. 
\*=========================================*/
{ 
   sigset_t ens_sig; 
   
   sigemptyset(&ens_sig); 
   sigprocmask(SIG_SETMASK, &ens_sig, NULL); 
}

/*
 #####                  #######
#     #   #   #   ####     #        #    #    #  ######  #####
#          # #   #         #        #    ##  ##  #       #    #
 #####      #     ####     #        #    # ## #  #####   #    #
      #     #         #    #        #    #    #  #       #####
#     #     #    #    #    #        #    #    #  #       #   #
 #####      #     ####     #        #    #    #  ######  #    #
*/

void InitSysTimer(struct itimerval *SysTimer)
/*=========================================*\
  Initialise la variable SysTimer a une 
  valeur nulle (qui desarmerait le timer
  s'il etait effectivement applique au 
  timer du systeme.)
\*=========================================*/
{ 
  SysTimer->it_interval.tv_sec  = 0; 
  SysTimer->it_interval.tv_usec = 0;
  
  SysTimer->it_value.tv_sec  = 0; 
  SysTimer->it_value.tv_usec = 0; 
}


void StopSysTimer()
/*=========================================*\
  Desarme le timer systeme de facon a ce que
  plus aucun signal SIGALRM ne sera delivre. 
\*=========================================*/
{
  struct itimerval SysTimer;

  SysTimer.it_interval.tv_sec  = 0; 
  SysTimer.it_interval.tv_usec = 0;
  
  SysTimer.it_value.tv_sec  = 0; 
  SysTimer.it_value.tv_usec = 0; 

  setitimer(ITIMER_REAL, &SysTimer, NULL); 
}


void SetSysTimerValue(float duree)
/*=========================================*\
  Affecte duree a la valeur du timer du 
  systeme. 
\*=========================================*/
{
  float value; 
  struct itimerval SysTimer; 
  long sec, microsec; 

  sec = (long)duree; 
  microsec = (long)(1000000*(duree - (long)duree)); 

  InitSysTimer(&SysTimer); 
  
  SysTimer.it_value.tv_sec  = (long)duree; 
  SysTimer.it_value.tv_usec = 
    (long)(1000000*(duree - (long)duree)); 
  
  setitimer(ITIMER_REAL, &SysTimer, NULL); 
}


float GetSysTimerValue()
/*=========================================*\
   Donne (en secondes) la duree jusqu'a 
   expiration du timeout et de l'emission
   d'un signal SIGALRM.
\*=========================================*/
{
  float value;
  struct itimerval SysTimer; 
  
  getitimer(ITIMER_REAL, &SysTimer);
  
  value = (float)(SysTimer.it_value.tv_sec
    + SysTimer.it_value.tv_usec/1000000.);

  return(value);
}


void WakeUp()
/*=========================================*\
   Lorsqu'on utilise la fonction sleep et 
   peut-etre quelques autres fonctions 
   dangereuses, le timer systeme peut 
   expirer, sans que le signal SIGALRM soit 
   pris en compte. Le probleme est qu'ensuite,
   le timer systeme ne bouge plus. Donc 
   de temps en temps il faut le surveiller 
   et le relancer s'il s'est endormi. 
\*=========================================*/
{
   if((FirstTimer)&&(GetSysTimerValue()==0))
	TimeoutHandler();
}


/*
   #                                    #######
  # #    #####   #####   #          #      #        #    #    #  ######  #####
 #   #   #    #  #    #  #          #      #        #    ##  ##  #       #    #
#     #  #    #  #    #  #          #      #        #    # ## #  #####   #    #
#######  #####   #####   #          #      #        #    #    #  #       #####
#     #  #       #       #          #      #        #    #    #  #       #   #
#     #  #       #       ######     #      #        #    #    #  ######  #    #
*/

AppliTimer_t* NewAppliTimer(int identite, float duree)
/*=========================================*\
  Alloue un nouveau timer appli et initialise
  sa duree a "duree". 
\*=========================================*/
{ 
  AppliTimer_t *timer; 

  timer = (AppliTimer_t*)malloc(sizeof(AppliTimer_t));

  if(timer==NULL)
    {fprintf(stderr,"NewTimer : malloc a echoue.\n");
    exit(1);}

  else
    {
      timer->identite = identite; 
      timer->duree    = duree; 
      timer->next     = NULL; 
    }
  return(timer);
}


void InitTimers(int TPipe)
/*=========================================*\
  Initialise la liste des timers de notre
  application. Au depart cette liste est 
  vide. 
\*=========================================*/
{ 
  struct itimerval SysTimer; 

  InitSysTimer(&SysTimer);

  p_action.sa_handler = &TimeoutHandler;
  sigaction(SIGALRM, &p_action, NULL); 
  
  FirstTimer=NULL; 
  
  TimerPipe = TPipe; 
}


void UpdateAllAppliTimers()
/*=========================================*\
  Lit la valeur de SysTimer. SysTimer a ete
  arme avec le premier timer de la liste.
  Faire la difference entre SysTimer et 
  ce temps et remplacer toutes les durees
  de tous les timers par la valeur prect
  - FirstTimer->duree + SysTimer. 
\*=========================================*/
{
  float offset; 
  AppliTimer_t *timer; 
  
  if(FirstTimer)
    {
      offset = FirstTimer->duree - GetSysTimerValue(); 
      
      FOR_ALL_TIMERS_IN_LIST(timer) 
	timer->duree -= offset; 
    }
}



void RemoveFirstAppliTimer()
/*=========================================*\
  Remplace le premier timer par le deuxieme
  et elimine le premier timer. 
\*=========================================*/
{
  AppliTimer_t *Morituri;

  Morituri = FirstTimer; 
  FirstTimer = FirstTimer->next; 
  free(Morituri);
}


void InsertFirstAppliTimer(AppliTimer_t *timer)
/*=========================================*\
  timer sera place au premier rang.
\*=========================================*/
{
  struct itimerval *Morituri;

  timer->next = FirstTimer; 
  FirstTimer  = timer; 
}


void InsertAppliTimer(AppliTimer_t *timer, AppliTimer_t *prec)
/*=========================================*\
  timer sera insere juste devant prec. 
  En d'autre termes, timer doit etre un peu
  plus urgent que prec. 
\*=========================================*/
{
  timer->next = prec->next; 
  prec->next  = timer; 
}


void RemoveAppliTimer(AppliTimer_t *RemovedTimer)
/*=========================================*\
  RemovedTimer sera elimine de la liste des 
  timers. 
\*=========================================*/
{
  AppliTimer_t *timer, *prev_timer; 

  if(RemovedTimer==FirstTimer)
    RemoveFirstAppliTimer();

  else
    {
      FOR_ALL_TIMERS_IN_LIST(timer) 
	if(timer->next == RemovedTimer)
	  prev_timer = timer; 
      
      /* prev_timer est le timer juste avant RemovedTimer */
      
      prev_timer->next = RemovedTimer->next; 
    }

  free(RemovedTimer); 
}


void ClassifyAppliTimerInList(AppliTimer_t *NewTimer)
/*=========================================*\
  Place le nouveau timer derriere ceux qui 
  sont plus urgents et devant ceux qui le 
  sont moins. 
\*=========================================*/
{
  AppliTimer_t *timer; 
  boolean classified=FAUX; 

/* ----------Insertion au premier rang ----------- */

  if((FirstTimer==NULL)||
     (NewTimer->duree < FirstTimer->duree))
    {
      InsertFirstAppliTimer(NewTimer);
      classified=VRAI;
    }

  else
    {
      FOR_ALL_TIMERS_IN_LIST_WHILE(timer, classified==FAUX) 
	{
/* --------- Insertion au dernier rang ------------*/       

	if(((timer->next==NULL)&&    
	    (NewTimer->duree >= timer->duree)) ||

/* -------------Les autres cas --------------------*/

	   ((timer->next!=NULL)&&
	    (NewTimer->duree >= timer->duree)&&
	    (NewTimer->duree <  timer->next->duree)))
	  {
	    InsertAppliTimer(NewTimer, timer); 
	    classified=VRAI;
	  }
	}
    }
}


void CheckTimers(char *comment)
/*=========================================*\
  Affiche a l'ecran tous les timers armes 
  avec leurs caracteristiques. 
\*=========================================*/
{
  AppliTimer_t *timer; 

  WakeUp();
  fprintf(stderr,"\nCHECK-TIMERS : %s\n", comment);
  fprintf(stderr,"SysTimer : \t a %f secondes de l'expiration.\n", GetSysTimerValue());
  FOR_ALL_TIMERS_IN_LIST(timer) 
    fprintf(stderr,"Timer %d \t a %f secondes de l'expiration.\n",
	    timer->identite, timer->duree);

  fprintf(stderr,"\n");
}


void StartATimer(int identite, float duree)
/*=========================================*\
  Remettre a jour toutes les durees de tous 
  les temporisateurs. 
  Si duree est plus petite que toutes les 
  autres, la mettre en tete et changer la 
  valeur de SysTimer. 
  Sinon classer le nouveau timer parmi les 
  autres dans l'ordre des durees. 
\*=========================================*/
{
  AppliTimer_t *timer, *NewTimer; 
  boolean impossible=FAUX; 

  WakeUp();

/*--------- Voir si le timer n'existe pas deja ----------*/

  FOR_ALL_TIMERS_IN_LIST(timer)
    if(timer->identite == identite)
      {fprintf(stderr,"Le timer %d existe deja.\n",identite);
      impossible=VRAI;}

/*--------- Valeurs de parametre incorrectes ------------*/

  if((!impossible)&&(identite<=0))
      {fprintf(stderr,"StartATimer : L'identite doit etre strictement positive.\n");
      fprintf(stderr,"La valeur de %d n'est pas autorisee.\n", identite);
      impossible=VRAI;}

  if((!impossible)&&(duree<=0))
      {fprintf(stderr,"StartATimer : La duree doit etre strictement positive.\n");
      fprintf(stderr,"La valeur %f donnee pour le timer %d n'est pas autorisee.\n",
	      duree, identite);
      impossible=VRAI;}

/*--------- Allocation et insertion dans la liste---------*/

  if(!impossible)
    {
      NewTimer = NewAppliTimer(identite, duree); 
     
      UpdateAllAppliTimers();
      
      if((FirstTimer==NULL)||
	 ((FirstTimer!=NULL)&&(duree < FirstTimer->duree)))	
	SetSysTimerValue(duree);
      
      ClassifyAppliTimerInList(NewTimer);
    }
}



void StopTimer(int identite)
/*=========================================*\
  Enleve le timer identite de la liste. 
\*=========================================*/
{
  AppliTimer_t *timer; 
  boolean trouve=FAUX; 

  WakeUp();

  if(FirstTimer->identite == identite)
    {
      if(FirstTimer->next!=NULL)
	SetSysTimerValue(FirstTimer->next->duree);
      else
	StopSysTimer();

      RemoveFirstAppliTimer();
      trouve=VRAI; 
    }

  else
    FOR_ALL_TIMERS_IN_LIST(timer) 
      if(timer->identite == identite)
	{RemoveAppliTimer(timer);  trouve=VRAI;}

  if(trouve==FAUX)
    fprintf(stderr,"StopTimer : le timer %d n'existe pas.\n",
	    identite);
}



int GetNextExpiredTimerID(int PipeDesc)
/*=========================================*\
  S'il y a un nombre dans le tube PipeDesc
  alors cette fonction rend sa valeur et 
  l'enleve du fifo, sinon, la fonction attend
  qu'il y ait effectivement un nombre sur le 
  tube. 
\*=========================================*/
{
  int res, TimerID=0; 

  WakeUp();

  res = read(PipeDesc, &TimerID, sizeof(int)); 

  if(res==CRASH) 
    {perror("GetNextExpiredTimerID : read");
    return(CRASH);}

  else return(TimerID);
}


void TimeoutHandler()
/*=========================================*\
  Il faut enlever le permier timer
  de la liste. Mettre son id dans TimerPipe
  et diminuer tous les temporisations des 
  autres Timer_t de la duree initiale du 
  timer qu'on vient d'ecarter. 
  Rearmer le temporisateur avec la duree 
  du nouveau premier timer.  
\*=========================================*/
{
  static int niv_recurs=0; 
  int res; 
  float Duree;
   
  MaskSigAlarm();

  res = write(TimerPipe, &(FirstTimer->identite), sizeof(int)); 
  if(res==-1) perror("ReactionAExpiration write");

  UpdateAllAppliTimers();
  
  if(FirstTimer->next!=NULL)
    {
      if(FirstTimer->next->duree > 0)
	{
	  SetSysTimerValue(FirstTimer->next->duree);
	  RemoveFirstAppliTimer();
	}
      else
	{
	  RemoveFirstAppliTimer();
	  niv_recurs++; 
	  TimeoutHandler();
	  niv_recurs--; 
	}
    }
  else
    {
      RemoveFirstAppliTimer();
      StopSysTimer();
    }
  
  if(niv_recurs==0)
    {
      p_action.sa_handler = &TimeoutHandler;
      sigaction(SIGALRM, &p_action, NULL);
      UnmaskSigAlarm();
    } 
}
