
//================================================
//
//	Lundi 23 novembre 1998
//
//	Arash HABIBI
//
//	Physique.C
//
//      Tout ce qui concerne la couche physique
//
//================================================


#include "Physique.H"


CouchePhys_c::CouchePhys_c(char *loc, char *dist, int data_p2l, int data_l2p, int disc_p2l, int disc_l2p)
//================================================
// Initialise les quelques variables et commence
// la boucle de la couche physique qui attend que 
// les evenements aient lieu. 
//================================================
{
  _Data_Phys_Lia_pipe =  data_p2l; 
  _Data_Lia_Phys_pipe =  data_l2p; 
  _Deconnexion_Phys_Lia_pipe =  disc_p2l;
  _Deconnexion_Lia_Phys_pipe =  disc_l2p;

  strcpy(_LocalSocketName, loc);
  strcpy(_DistantSocketName, dist);

  _stop = FAUX; 

  _SetLocalAddress(loc);
  _SetDistantAddress(dist); 

  _Loop();
}


CouchePhys_c::~CouchePhys_c()
//================================================
// ferme la socket et efface les fichiers 
// temporaires qui ont ete crees. 
//================================================
{
  printf("Couche physique deconnectee.\n");
  close(_sock_desc);
  remove(_LocalSocketName);
}



void CouchePhys_c::_SetLocalAddress(char *ref)
//================================================
// l'equivalent de GetAndBindSocket. Ouvre une 
// socket et l'attache a l'adresse ref.
//================================================
{
    _sock_desc = socket(AF_UNIX, SOCK_DGRAM, 0);
    
    if(_sock_desc == -1) {perror("socket"); exit(1);}
    
// ----- preparation de l'adresse locale ----- 
    
    _adresse_loc.sun_family = AF_UNIX; 
    strcpy(_adresse_loc.sun_path, ref);

// ----- et attachement de la socket a   ----- 
// ----- l'adresse ref par bind          ----- 
	
    if(bind(_sock_desc, (struct sockaddr*) &_adresse_loc, 
	       sizeof(_adresse_loc))==-1)
      {perror(ref); exit(1);}
}




void CouchePhys_c::_SetDistantAddress(char *ref)
//================================================
// Caclule simplement l'adresse de la socket 
// distante
//================================================
{
  _adresse_dist.sun_family = AF_UNIX; 
  strcpy(_adresse_dist.sun_path, ref);    
}


void CouchePhys_c::_Disconnect()
//================================================
// Cette fonction est appelee quand la couche 
// physique a recu un message de la part de 
// la couche liaison sur le SAP Deconnexion
// Dans ce cas, il ne peut s'agir que d'une 
// demande de deconnexion. C'est pourquoi on 
// envoie une trame vide qui ne contient que 
// le caractere '.'
//================================================
{
  if(sendto(_sock_desc, ".", 1, 0, 
	    (struct sockaddr*) &_adresse_dist, 
	    TAILLE_ADRESSE)==-1)
    {perror("_Disconnect sendto"); exit(1);}
}


void CouchePhys_c::_SendData()
//================================================
// On vient de recevoir des donnees de la part
// de la couche liaison sur le SAP Data. Donc 
// il ne reste plus qu'a l'envoyer sur la 
// socket. 
//================================================
{
  int taille_trame, taille_adresse=TAILLE_ADRESSE; 
  char trame[TAILLE_TRAME];

  if((taille_trame = read(_Data_Lia_Phys_pipe, trame, 
			  TAILLE_TRAME))<=0)
    {perror("_SendData read"); exit(1);}

  if(sendto(_sock_desc, trame, taille_trame, 0, 
	    (struct sockaddr*) &_adresse_dist, 
	    TAILLE_ADRESSE)==-1)
    {perror("_SendData sendto"); exit(1);}
}



void CouchePhys_c::_ReceiveData()
//================================================
// On vient de recevoir des donnees sur la socket
// il ne reste plus qu'a les envoyer a la 
// couche liaison sur le SAP Data. 
//================================================
{
  int taille_trame, taille_adresse=TAILLE_ADRESSE; 
  char trame[TAILLE_TRAME], disconnect[1];

  if((taille_trame = recvfrom(_sock_desc, trame, TAILLE_TRAME, 0, 
			  (struct sockaddr *) &_adresse_dist,
			  &taille_adresse))<=0)
    {perror("Phys : recvfrom"); exit(1);}

  if((taille_trame==1)&&(trame[0]=='.'))
    {
      _stop = VRAI; 
      if(write(_Deconnexion_Phys_Lia_pipe, disconnect, 1)==-1)
	{perror("_ReceiveData : write1"); exit(1);}
    }

  else if(taille_trame>0)
    if(write(_Data_Phys_Lia_pipe, trame, taille_trame)==-1)
      {perror("_ReceiveData : write2"); exit(1);}
}



void CouchePhys_c::_Loop()
//================================================
// la grande boucle de la couche physique qui 
// attend les evenements qui peuvent se produire : 
// il y a trois evenements possible : 
// - arrive d'une trame sur la socket
// - arrive d'un message de la part de la couche liaison
// - arrive d'une demande de deconnexion venant de 
//   la couche liaison.
//================================================
{
    fd_set lecture, ecriture, except;
    int nbfds, nb_events;

    nbfds = Max(_sock_desc, _Data_Lia_Phys_pipe, _Deconnexion_Lia_Phys_pipe) + 1;

    printf("La couche physique commence a boucler.\n");
    
    while(!_stop)
	{
	  //---- Preparation des ensembles -------- 
	  //--------- de descripteurs ------------- 
	    
	    FD_ZERO(&lecture); 
	    FD_ZERO(&ecriture); 
	    FD_ZERO(&except); 
	    FD_SET(_sock_desc, &lecture); 
	    FD_SET(_Data_Lia_Phys_pipe, &lecture); 
	    FD_SET(_Deconnexion_Lia_Phys_pipe, &lecture); 

	    
	    //------- attente d'un evenment --------- 

	    nb_events = select (nbfds, &lecture, &ecriture, 
				&except, NULL);
	    if(nb_events==-1) {perror("select"); exit(1);} 
	    
	    //------- le depouillement des votes ---- 
	    
	    if(FD_ISSET(_sock_desc, &lecture))
	      _ReceiveData();

	    
	    if(FD_ISSET(_Data_Lia_Phys_pipe, &lecture))
	      _SendData();
	    
	    else if(FD_ISSET(_Deconnexion_Lia_Phys_pipe, &lecture))
		{_Disconnect(); _stop = VRAI;}
	}
}



