
/*================================================*\

	Mardi 24 novembre 1998

	Arash HABIBI

	Attente.c

	A chaque fois qu'il faut attendre un 
	evenement c'est ici qu'il faut s'adresser. 

\*================================================*/


#include "Attente.h"


int Attendre(int fd1, int fd2, int fd3, int fd4, float timeout)
/*================================================*\
  Attend que l'un des descripteurs fd1, fd2, fd3 
  ou fd4 deviennent lisibles ou alors que le temps 
  timeout soit ecoule. Si timeout=0 alors le 
  temporisateur ne sera pas arme. De meme si l'un 
  des descripteurs est negatif alors l'argument sera 
  ignore. 

  La fonction rend YaALireSur1___ (i.e. 1) si fd1 est seul lisible
                   YaALireSur_2__ (i.e. 2) si fd2 est seul lisible
                   YaALireSur__3_ (i.e. 4) si fd3 est seul lisible
                   YaALireSur___4 (i.e. 8) si fd4 est seul lisible
		   YaALireSur12__ (i.e. 3) si fd1 et fd2 sont lisibles
		   YaALireSur1_3_ (i.e. 5) si fd1 et fd3 sont lisibles
		   etc. 
		   TIMEOUT (i.e. 0) si aucun n'est lisible avant
		     une duree de timeout. 
		   Si timeout=0 alors le temporisateur n'est pas
		   amorce. 

  IMPORTANT : on travaille seulement en lecture,
  le descripteur fd4 (s'il est positif) est examine 
  SEULEMENT si les descripteur fd1, fd2, fd3 ne sont 
  pas lisibles ou sont negatifs. 

\*================================================*/
{
    fd_set lecture, ecriture, except;
    int nbfds, nb_events, res; 
    boolean crash=FAUX, stop=FAUX;
    struct timeval timer;

    /*---- Preparation du timer -------- */

    timer.tv_sec  = (int)floor(timeout); 
    timer.tv_usec = (int)floor(1000000*(timeout-(int)floor(timeout)));

    /*---- Preparation des ensembles -------- */
    /*--------- de descripteurs ------------- */
    
    nbfds = Max(fd1, fd2, fd3, fd4) + 1; 
    if(nbfds<=0) nbfds=1; 
    
    FD_ZERO(&lecture); 
    FD_ZERO(&ecriture); 
    FD_ZERO(&except); 
    if(fd1>=0) FD_SET(fd1, &lecture); 
    if(fd2>=0) FD_SET(fd2, &lecture); 
    if(fd3>=0) FD_SET(fd3, &lecture); 
    if(fd4>=0) FD_SET(fd4, &lecture); 
    
    /*------- attente d'un evenment --------- */
    
    if(timeout==0)
      nb_events = select (nbfds, &lecture, &ecriture, 
			  &except, NULL);
    else
      nb_events = select (nbfds, &lecture, &ecriture, 
			  &except, &timer);

    if(nb_events==-1) {perror("select"); return(-1);} 
    
    /*------- le depouillement des votes ---- */
    
    else
      {
	res = 0; 

	if((fd1>=0)&&(FD_ISSET(fd1, &lecture)))
	  res+=1; 

	if((fd2>=0)&&(FD_ISSET(fd2, &lecture)))
	  res+=2; 

	if((fd3>=0)&&(FD_ISSET(fd3, &lecture)))
	  res+=4; 

	if(((fd1<0)||(!(FD_ISSET(fd1, &lecture))))&&
	   ((fd2<0)||(!(FD_ISSET(fd2, &lecture))))&&
	   ((fd3<0)||(!(FD_ISSET(fd3, &lecture)))))
	  if((fd4>=0)&&(FD_ISSET(fd4, &lecture)))
	    res+=8; 

	return(res);
      }
}



