
/*================================================*\

	Lundi 23 novembre 1998

	Arash HABIBI

	Physique.c

      Tout ce qui concerne la couche physique

\*================================================*/


#include "Physique.h"


void InitCouchePhys(CouchePhys_t *CchPhys, int id, int data_pl, int data_lp, int disc_pl, int disc_lp)
/*================================================*\
 Initialise les quelques variables (essentiellement
 les descripteurs de tube entre la couche physique et
 la couche liaison, et ensuite commence la boucle de 
 la couche physique qui attend que les evenements 
 aient lieu. 
\*================================================*/
{
  fprintf(stderr, "Couche Physique %d : commence.\n", id);
  CchPhys->Data_PL =  data_pl; 
  CchPhys->Data_LP =  data_lp; 
  CchPhys->Disc_PL =  disc_pl;
  CchPhys->Disc_LP =  disc_lp;

  CchPhys->identite = id; 

  if(id==STATION1) 
    {
      SetLocAdrSock("sock_station1", CchPhys);
      SetDstAdrSock("sock_medium1", CchPhys); 
    }
  else /* (id==STATION2) */
    {
      SetLocAdrSock("sock_station2", CchPhys);
      SetDstAdrSock("sock_medium2", CchPhys); 
    }

  Phys_Boucle(CchPhys);
}


void FreeCouchePhys(CouchePhys_t *CchPhys)
/*================================================*\
 ferme la socket et efface les fichiers 
 temporaires qui ont ete crees. 
\*================================================*/
{
  close(CchPhys->desc_sock);

  if(CchPhys->identite==STATION1)
    remove("sock_station1");

  else /* (CchAPhys->identite==STATION2) */
    remove("sock_station2");

  fprintf(stderr,"C'est la fin de la couche physique.\n");
}


void SetLocAdrSock(char *ref, CouchePhys_t *CchPhys)
/*================================================*\
 l'equivalent de GetAndBindSocket. Ouvre une 
 socket et l'attache a l'adresse ref.
\*================================================*/
{
  CchPhys->desc_sock = socket(AF_UNIX, SOCK_DGRAM, 0);
    
    if(CchPhys->desc_sock == -1) {perror("socket"); exit(1);}
    
    /* ----- preparation de l'adresse locale ----- */
    
    CchPhys->adresse_loc.sun_family = AF_UNIX; 
    strcpy(CchPhys->adresse_loc.sun_path, ref);

    /*----- et attachement de la socket a   ----- */
    /*----- l'adresse ref par bind          ----- */
	
    if(bind(CchPhys->desc_sock, 
	    (struct sockaddr*) &(CchPhys->adresse_loc),
	       sizeof(CchPhys->adresse_loc))==-1)
      {perror(ref); exit(1);}
}




void SetDstAdrSock(char *ref, CouchePhys_t *CchPhys)
/*================================================*\
 Caclule simplement l'adresse de la socket 
 distante
\*================================================*/
{
  CchPhys->adresse_dist.sun_family = AF_UNIX; 
  strcpy(CchPhys->adresse_dist.sun_path, ref);    
}


void Phys_ViderTubeInterface(int tube)
/*================================================*\
  Certains tubes n'ont pas besoin d'etre vraiment
  "lues" car leur contenu n'a pas beaucoup 
  d'importance. C'est le fait qu'il y ait un 
  message qui est important. C'est tout ce qui est 
  tube d'interface (disc, conn etc). Or si on ne 
  les lit pas, il faut tout de meme les vider, 
  sinon, la prochaine fois qu'on lira dessus, 
  les anciennes donnees y seront toujours. 
\*================================================*/
{
  char message[TAILLE_ASPI];

  if(read(tube, message, TAILLE_ASPI)<=0)
    {perror("Phys_ViderTubeInterface read"); exit(1);}
}


void Phys_DisconnectIndication(int Disc_PL, int Disc_LP)
/*================================================*\
  Pour dire a la couche liaison qu'une deconnexion
  a ete demandee sur la ligne par l'entite paire. 
  Et attendre que la couche liaison ait aussi 
  termine de lire le contenu des tubes. 
\*================================================*/
{
  /*--- Envoyer l'indication de --------------------*/
  /*----deconnexion a la couche liaison ------------*/

  if(write(Disc_PL, INDICATION, 1)==-1)
    {perror("Phys_DisconnectIndication : write"); 
    exit(1);}
  
  /*--- attendre que la couche liaison ait termine aussi -------*/
  /*--- avant de lui retirer les sockets et les tubes ----------*/

  if(read(Disc_LP, RESPOND, 1)==-1)
    {perror("Phys_DisconnectIndication : read"); exit(1);}
}


int Phys_SendData(int desc_sock, struct sockaddr_un adr_dist, int Data_LP)
/*================================================*\
  On vient de recevoir des donnees de la part
  de la couche liaison sur le SAP Data. Donc 
  il ne reste plus qu'a l'envoyer sur la 
  socket. On retourne 1 pour dire que tout 
  s'est bien passe, et -1 pour 
  dire qu'il y a eu un probleme. 
\*================================================*/
{
  int taille_trame, taille_adresse=TAILLE_ADRESSE; 
  char trame[TAILLE_TRAME];
  
  if((taille_trame = read(Data_LP, 
			  trame, 
			  TAILLE_TRAME))<=0)
    {perror("Phys_SendData read"); return(-1);}

  else if(sendto(desc_sock, trame, taille_trame, 0, 
	    (struct sockaddr*) &adr_dist, 
	    TAILLE_ADRESSE)==-1)
    {perror("Phys_SendData sendto"); return(-1);}

  else
    return(1);
}


int Phys_ReceiveData(int desc_sock, struct sockaddr_un adr_dist, int Data_PL)
/*================================================*\
 On vient de recevoir des donnees sur la socket
 il ne reste plus qu'a les envoyer a la 
 couche liaison sur le SAP Data. 
 La fonction rend 1 si tout va bien et -1 s'il 
 y a un probleme. Elle rend 0 si l'entite 
 physique paire indique que la ligne va bientot
 etre rompue. 
\*================================================*/
{
  int taille_trame, taille_adresse=TAILLE_ADRESSE; 
  char trame[TAILLE_TRAME];

  /*--- Lire ce qui se trouve sur la socket ------------*/

  if((taille_trame = recvfrom(desc_sock, 
			      trame, TAILLE_TRAME, 0, 
			      (struct sockaddr *) &adr_dist,
			      &taille_adresse))<=0)
    {perror("Phys : recvfrom"); return(-1);}

  /*--- Ce qui suit est une indication de deconnexion---*/
  /*--- de l'entite physique paire ---------------------*/

  else if((taille_trame==1)&&(trame[0]=='.'))
    return(0);

  /*--- On a recu de l'information ordinaire qu'on ---- */
  /*--- delivre comme data a la couche superieure ----- */

  else if(taille_trame>0)
    if(write(Data_PL, trame, taille_trame)==-1)
      {perror("Phys_ReceiveData : write2"); return(-1);}
    else
      return(1);
}



int Phys_Quit(int desc_sock, struct sockaddr_un adr_dist)
/*================================================*\
  Cette fonction est appelee quand la couche 
  physique a recu un message de la part de 
  la couche liaison sur le SAP Deconnexion
  Dans ce cas, il ne peut s'agir que d'une 
  demande de deconnexion. C'est pourquoi on 
  envoie une trame vide qui ne contient que 
  le caractere '.'
  On retourne 1 pour dire que tout s'est 
  bien passe. 
\*================================================*/
{
  if(sendto(desc_sock, ".", 1, 0, 
	    (struct sockaddr*) &adr_dist, 
	    TAILLE_ADRESSE)==-1)
    {perror("Phys_Quit sendto"); return(-1);}
  else
    return(1);
}


void Phys_Boucle(CouchePhys_t *CchPhys)
/*================================================*\
 la grande boucle de la couche physique qui 
 attend les evenements qui peuvent se produire : 
 il y a trois evenements possible : 
 - arrive d'une trame sur la socket
 - arrive d'un message de la part de la couche liaison
 - arrive d'une demande de deconnexion venant de 
   la couche liaison.
\*================================================*/
{
    fd_set lecture, ecriture, except;
    int res_attente, res; 
    boolean crash=FAUX, stop=FAUX;
    
    while((stop==FAUX)&&(crash==FAUX))
	{
	  res_attente = Attendre(CchPhys->desc_sock, 
			       CchPhys->Data_LP, 
			       -1, 
			       CchPhys->Disc_LP, 
			       0.0);

	  if(res_attente==-1) crash=VRAI; 

	  else 
	    {
/*--------------------- S'il y a quelque chose a lire sur la socket ---- */

	      if((res_attente==YaALireSur1___)||(res_attente==YaALireSur12__)||
		 (res_attente==YaALireSur1_3_)||(res_attente==YaALireSur123_))
		{
		  res = Phys_ReceiveData(CchPhys->desc_sock, 
					 CchPhys->adresse_dist, 
					 CchPhys->Data_PL);
		  if(res<0) crash=VRAI; 
		  else if(res==0) /* l'entite physique paire a demande une disc */
		    {
		      stop=VRAI; 
		      Phys_DisconnectIndication(CchPhys->Disc_PL, 
						CchPhys->Disc_LP);
		    }
		  else /* tout va bien */;
		}

/*-------------------------- Si la couche liaison envoie des donnees ---- */
	      
	      if((res_attente==YaALireSur_2__)||(res_attente==YaALireSur12__)||
		 (res_attente==YaALireSur_23_)||(res_attente==YaALireSur123_))
		{
		  res = Phys_SendData(CchPhys->desc_sock, 
				      CchPhys->adresse_dist, 
				      CchPhys->Data_LP);
		  if(res<0) crash=VRAI; 
		  else /* tout va bien */;
		}

/*--------------------- Si la couche liaison demande une deconnexion -----*/
	      
	      if(res_attente==YaALireSur___4)
		{
		  Phys_ViderTubeInterface(CchPhys->Disc_LP);

		  res = Phys_Quit(CchPhys->desc_sock, 
				  CchPhys->adresse_dist);
		  if(res<0) crash=VRAI; 
		  else stop=VRAI;
		}
	    }
	}

    if(stop==VRAI) fprintf(stderr,"Couche Physique  : Echange termine.\n");
    else /*crash==VRAI*/ fprintf(stderr,"Couche Physique : Echange interrompu.\n");
    
    FreeCouchePhys(CchPhys);
    fprintf(stderr, "Couche Physique %d: deconnectee.\n", CchPhys->identite);
}



