
/*===============================================*\
	Vendredi 18 Septembre 1998

	Arash HABIBI

	ServeurTCP.c

	Un espece de talk simplifie. 
	Ce cote-ci joue le role de serveur, 
	simplement dans le sens ou c'est lui 
	qui ecoute le reseau en attendant une
	demande de connexion. 
	L'interet de la manip est de montrer 
	le maniement des primitives en mode
	non-bloquant. 

\*================================================*/

#define OK 1

#include <errno.h>

#include "SocketToolbox.h"


void ConnexionServeur(short port_loc, int *sock_ecoute, int *sock_service)
/*===============================================*\
\*===============================================*/
{
	struct sockaddr_in adresse_loc, adresse_dist;
	int res, taille_adr, nb_pendants=1; 
	AdresseIP_t IP; 
	u_short port_dist;
	
/*-------- Creation et attachement de la socket d'ecoute ---------*/
	
	*sock_ecoute = GetAndBindSocket(SOCK_STREAM, 
		port_loc, &adresse_loc);
	
	printf("PapotServTCP attend interlocuteur sur port numero %d\n",
		adresse_loc.sin_port);
		
		
/*-------- Debut de l'ecoute (Entree en service du serveur) ------*/
			
	res = listen(*sock_ecoute, nb_pendants);
	
	if(res == -1) {perror("listen"); exit(1);}
	else printf("PapotServTCP commence l'ecoute\n");
	
	
/*-------- La boucle de l'acceptation de la demande de cnxn -------*/

	taille_adr = sizeof(adresse_dist);

	*sock_service = accept(*sock_ecoute, 
		(struct sockaddr *)&adresse_dist, 
		&taille_adr);
	if(res==-1) {perror("accept");}
		
	/* --------connexion acceptee ------------*/

	ReadNetworkAddress(adresse_dist, &IP, &port_dist);
			
	printf("Demande de connexion venant de l'hote\
	     	\n%d.%d.%d.%d du port numero %d.\
	     	\nLa socket de service est le numero %d\n\n",
		IP.n1, IP.n2, IP.n3, IP.n4, port_dist, 
		*sock_service);
}
	


int GaverSocket(int desc_sock, char *message, int taille)
/*===============================================*\
  En mode non-bloquant, lorsqu'on ecrit dans une
  socket, tout n'est pas necessairement 
  transmis. Ce qui peut etre transmis est 
  transmis et le reste est ignore. Donc c'est
  nous qui devons tenir compte de cette partie 
  ignoree et la reexpedier des que possible. 
  Cette fonction rend 1 si le message courant 
  a ete transmis et qu'on peut aller chercher 
  un autre message. 
\*===============================================*/
{
    static int index=0; 
    static int transmis=0; 
    int res; 
    char *buffer;		/* la partie non-transmise */ 
				/* du message              */
    if(index==0) transmis=0; 
    
    buffer = message+index; 

    res = write(desc_sock, buffer, taille-index);
    if((res==-1)&&(errno!=EWOULDBLOCK))  
	{perror("GaverSocket"); return(-1);}

    /* else if ((res==-1)&&(errno==EWOULDBLOCK)); */
    /* il ne se passe rien ---------------------- */

    else if(res==0) return(-1); /* socket paire fermee */

    else if(res>0) 
	{
	    index += res; 
	    if(index==taille)
		{
		    index=0; 
		    transmis=1; 
		}
	}
    return(transmis); 
}
       
    
void main(int argc, char **argv)
/*===============================================*\
\*===============================================*/
{
	int term_rec, term_send, sock_ecoute, sock_service; 
	int transmission_over, crash, nonbloquant; 
	int nb_lu, nb_recu, res, msg_prcdt_transmis=OK; 
	char message[100];
	 
	if(argc!=2) 
	{printf("Usage : PapotServ <NomTerminal>\n"); exit(1);}

/*-------- Ouvertures diverses -------- ------------------*/

	/*-- Le terminal sur lequel on enverra -------*/

	term_send = open("/dev/tty", O_RDWR); 
	if(term_send == -1) 
	    {perror("terminal de controle"); exit(1);}

	/*-- Le terminal sur lequel on recevra -------*/

	term_rec = open(argv[1], O_RDWR); 
	if(term_rec == -1) 
	    {perror("terminal de lecture"); exit(1);}

/*-------- Connexion ------------------ ------------------*/
	
	ConnexionServeur(0, &sock_ecoute, &sock_service);
	transmission_over = 0; crash = 0; 

	
/*-------- Passage en mode non-bloquant ------------------*/

	nonbloquant=1; 
	ioctl(sock_service, FIONBIO, &nonbloquant);

	nonbloquant=1; 
	ioctl(term_send, FIONBIO, &nonbloquant);
	      
/*-------- La boucle d'attente d'evenement ---------------*/

	while((!crash)&&(!transmission_over))
	{
	    /*-------------- lecture de la socket --------*/

		nb_recu = read(sock_service, message, 10);

		if((nb_recu==-1)&&(errno!=EWOULDBLOCK)) 
		    {perror("read socket"); crash=1;}

		else if((nb_recu==-1)&&(errno==EWOULDBLOCK)) 
		    /* silence radio rien a faire */;

		else if(nb_recu==0) /* socket paire fermee */
		    crash=1; 

		else if((nb_recu==2)&&
			(strncmp(message, ".", nb_recu-1)==0))
		    transmission_over=1; 

		else if(nb_recu>0)
		    {
			/*----- ecriture terminal ------- */

			res=write(term_rec, message, nb_recu);
			/* celui-ci est reste bloquant */
			if(res==-1) perror("write terminal");
		    }

		/*-------------- lecture du terminal -----*/

		if(msg_prcdt_transmis==OK)
		    nb_lu = read(term_send, message, 10);

		if((nb_lu==-1)&&(errno!=EWOULDBLOCK))
		    {perror("read terminal de ctrl"); crash=1;}

		else if((nb_lu==-1)&&(errno==EWOULDBLOCK));
		    /* Rien a lire, c'est tout */

		else if(nb_lu>0)
		    {
			/*----- envoi sur socket ---------*/

			/*res=write(sock_service, message, nb_lu);*/
			msg_prcdt_transmis=
			    GaverSocket(sock_service, message, nb_lu); 

			if(msg_prcdt_transmis==-1)
			    crash=1; 

			else if(msg_prcdt_transmis==OK)
			    if((nb_lu==2)&&
			       (strncmp(message, ".", nb_lu-1)==0))
				transmission_over=1; 
			    else;
				/* tout va bien on  */
				/* passe a la suite */
		    }
	}

	close(sock_service);
	close(sock_ecoute);
	if(crash)
		printf("Echange interrompu\n\n");
	else if(transmission_over)
		printf("Echange termine. Deconnexion.\n\n");
}
